import os

import ycm_core

cwd = os.path.dirname(os.path.realpath(__file__))

FLAGS = [
    "-Wall",
    "-Weffc++",
    "-std=c++14",
    "-Werr",
    "-x",
    "c++",
    "-I",
    ".",
    "-I",
    cwd + "/fwdpy11/headers/fwdpp",
    "-I",
    cwd + "/fwdpy11/headers",
]


def FlagsForFile(filename):
    dirname = os.path.dirname(filename)
    flags = {"flags": FLAGS, "do_cache": True}
    return flags



import unittest

import numpy as np

import fwdpy11


class test_stopping_criterion_DiploidPopulation(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(1000, 1.0)
        p = {
            "nregions": [],  # No neutral mutations -- add them later!
            "gvalue": fwdpy11.Additive(2.0),
            "sregions": [fwdpy11.ExpS(0, 1, 1, -0.1)],
            "recregions": [fwdpy11.Region(0, 1, 1)],
            "rates": (0.0, 1e-3, 1e-3),
            # Keep mutations at frequency 1 in the pop if they affect fitness.
            "prune_selected": False,
            "demography": np.array([1000] * 10000, dtype=np.uint32),
        }
        self.params = fwdpy11.ModelParams(**p)

    def test_stop(self):
        rng = fwdpy11.GSLrng(42)

        def generation(pop, simplified):
            if pop.generation >= 50:
                return True
            return False

        fwdpy11.evolvets(rng, self.pop, self.params, 100, stopping_criterion=generation)
        self.assertEqual(self.pop.generation, 50)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#


def evolvets(
    rng,
    pop,
    params,
    simplification_interval,
    recorder=None,
    post_simplification_recorder=None,
    suppress_table_indexing=False,
    record_gvalue_matrix=False,
    stopping_criterion=None,
    track_mutation_counts=False,
    remove_extinct_variants=True,
    preserve_first_generation=False,
):
    """
    Evolve a population with tree sequence recording

    :param rng: random number generator
    :type rng: :class:`fwdpy11.GSLrng`
    :param pop: A population
    :type pop: :class:`fwdpy11.DiploidPopulation`
    :param params: simulation parameters
    :type params: :class:`fwdpy11.ModelParams`
    :param simplification_interval: Number of generations between simplifications.
    :type simplification_interval: int
    :param recorder: (None) A temporal sampler/data recorder.
    :type recorder: function
    :param post_simplification_recorder: (None) A temporal sampler
    :type post_simplification_recorder: function
    :param suppress_table_indexing: (False) Prevents edge table indexing until
                                    end of simulation
    :type suppress_table_indexing: bool
    :param record_gvalue_matrix: (False) Whether to record genetic values into
                                 :attr:`fwdpy11.PopulationBase.genetic_values`.
    :type record_gvalue_matrix: bool
    :param preserve_first_generation: (False) Whether to record generation 0 as
                                      ancient samples. Must be `True` for
                                      tree sequence "recapitation". See
                                      :ref:`finishwithmsprime`.
    :type preserve_first_generation: bool

    The recording of genetic values into :attr:`fwdpy11.PopulationBase.genetic_values`
    is suppressed by default.  First, it is redundant with
    :attr:`fwdpy11.DiploidMetadata.g` for the common case of mutational effects on a
    single trait. Second, we save some memory by not tracking these matrices.
    However, it is useful to track these data for some
    cases when simulating multivariate mutational effects (pleiotropy).

    For a detailed description of `post_simplification_recorder`,
    see :ref:`tstimeseries`.

    .. note::
        If recorder is None,
        then :class:`fwdpy11.NoAncientSamples` will be used.

        If post_simplification_recorder is None, then
        :class:`fwdpy11.RecordNothing` will be used.

    .. versionchanged:: 0.5.2

        Added post_simplification_recorder.

    .. versionchanged:: 0.7.1

        Added preserve_first_generation.

    """
    import warnings

    # Test parameters while suppressing warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Will throw exception if anything is wrong:
        params.validate()

    if recorder is None:
        from ._fwdpy11 import NoAncientSamples

        recorder = NoAncientSamples()

    if post_simplification_recorder is None:
        from ._fwdpy11 import RecordNothing

        post_simplification_recorder = RecordNothing()
        reset_treeseqs_after_simplify = False
    else:
        reset_treeseqs_after_simplify = True

    if stopping_criterion is None:
        from ._fwdpy11 import _no_stopping

        stopping_criterion = _no_stopping

    from ._fwdpy11 import MutationRegions
    from ._fwdpy11 import dispatch_create_GeneticMap
    from ._fwdpy11 import evolve_with_tree_sequences

    pneutral = 0.0
    if params.mutrate_n + params.mutrate_s > 0.0:
        pneutral = params.mutrate_n / (params.mutrate_n + params.mutrate_s)
    mm = MutationRegions.create(pneutral, params.nregions, params.sregions)
    rm = dispatch_create_GeneticMap(params.recrate, params.recregions)

    from ._fwdpy11 import SampleRecorder

    sr = SampleRecorder()
    from ._fwdpy11 import _dgvalue_pointer_vector

    gvpointers = _dgvalue_pointer_vector(params.gvalue)
    evolve_with_tree_sequences(
        rng,
        pop,
        sr,
        simplification_interval,
        params.demography,
        params.simlen,
        params.mutrate_n,
        params.mutrate_s,
        mm,
        rm,
        gvpointers,
        recorder,
        stopping_criterion,
        params.pself,
        params.prune_selected is False,
        suppress_table_indexing,
        record_gvalue_matrix,
        track_mutation_counts,
        remove_extinct_variants,
        reset_treeseqs_after_simplify,
        preserve_first_generation,
        post_simplification_recorder,
    )



import numpy as np

import fwdpy11
import gvalue_recorder

N = 1000

pdict = {
    "demography": np.array([N] * 100, dtype=np.uint32),
    "nregions": [],
    "sregions": [fwdpy11.GaussianS(0, 1, 1, 0.25)],
    "recregions": [fwdpy11.Region(0, 1, 1)],
    "rates": (0.0, 5e-3, 5e-3),
    "gvalue": fwdpy11.Additive(2.0, fwdpy11.GSS(VS=1, opt=1)),
    "prune_selected": False,
}

params = fwdpy11.ModelParams(**pdict)

pop = fwdpy11.DiploidPopulation(N, 1.0)

rng = fwdpy11.GSLrng(42)

gvalues = []
r = gvalue_recorder.record_gvalue(gvalues)
fwdpy11.evolvets(rng, pop, params, 100, r)

assert len(gvalues) == pop.generation, "Callback failure"



import sqlite3

import pandas as pd

with sqlite3.connect("stats.sqlite3") as conn:
    df = pd.read_sql("select * from data", conn)

with sqlite3.connect("temp/stats.sqlite3") as conn:
    df2 = pd.read_sql("select * from data", conn)

print(df.head())
print(df2.head())
print(df.tail())
print(df2.tail())
print(df.equals(df2))



import pickle
import unittest

import numpy as np

import fwdpy11 as fp11
import fwdpy11.ezparams
import snowdrift


class SamplePhenotypes(object):
    """
    Temporal sampler checks that one can hook
    a stateful fitness model to a sampler
    and access its data and that the data
    are as expected.
    """

    def __init__(self, f):
        self.f = f
        # self.a = fwdpy11.genetic_values.DiploidAdditive(2.0)

    def __call__(self, pop):
        pass
        # for i in range(pop.N):
        #     w = self.a(pop.diploids[i], pop)
        #     assert(w == self.f.phenotypes[i])


def evolve_snowdrift(args):
    """
    We write the function taking a tuple
    out of habit, simplifying later
    integration with multiprocessing or
    concurrent.futures.
    """
    N, seed = args
    # Construct as single-deme object
    # with N diploids
    pop = fp11.DiploidPopulation(N)
    # Initialize a random number generator
    rng = fp11.GSLrng(seed)
    p = {
        "sregions": [fp11.ExpS(0, 1, 1, -0.1, 1.0)],
        "recregions": [fp11.Region(0, 1, 1)],
        "nregions": [],
        "gvalue": snowdrift.DiploidSnowdrift(0.2, -0.2, 1, -2),
        # evolve for 100 generations so that unit tests are
        # fast
        "demography": np.array([N] * 100, dtype=np.uint32),
        "rates": (0.0, 0.0025, 0.001),
        "prune_selected": False,
    }
    params = fwdpy11.ModelParams(**p)
    sampler = SamplePhenotypes(params.gvalue)
    fp11.evolve_genomes(rng, pop, params, sampler)
    # return our pop
    return pop


class testSnowdrift(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.f = snowdrift.DiploidSnowdrift(1, -1, 0.1, 0.2)

    def testShape(self):
        s = self.f.shape
        self.assertEqual(len(s), 1)
        self.assertEqual(s[0], 1)

    def test_genetic_values(self):
        self.assertEqual(len(self.f.genetic_values), 1)

    def testPickle(self):
        self.f.phenotypes = [1, 2, 3, 4]
        p = pickle.dumps(self.f, -1)
        up = pickle.loads(p)
        self.assertEqual(up.phenotypes, self.f.phenotypes)

    def test_evolve(self):
        p = evolve_snowdrift((1000, 42))
        p


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
# Test behavior of fixations in simulations.
# As of 0.1.3a1, fixations are to be sorted by position
# for all types of simulations, allowing for easier lookup.
#
# In 0.2.0, this unit test was refactored to skip simulation.
# Instead, we use pybind11 + cppimport to expose the underlying
# C++ API to Python, create pops manually, and test the API.

import os
import unittest

import fixation_properties as fp
import fwdpy11


class testFixationsAreSortedDiploidPopulation(unittest.TestCase):
    @classmethod
    def setUp(self):
        mutations = fwdpy11.MutationVector()
        fixations = fwdpy11.MutationVector()
        gametes = fwdpy11.HaploidGenomeVector()
        diploids = fwdpy11.DiploidVector()
        mutations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        mutations.append(fwdpy11.Mutation(0.2, 0.0, 1.0, 0, 0))
        gametes.append(fwdpy11.HaploidGenome((4, [1], [0])))
        diploids.append(fwdpy11.DiploidGenotype(0, 0))
        diploids.append(fwdpy11.DiploidGenotype(0, 0))
        self.pop = fwdpy11.DiploidPopulation(diploids, gametes, mutations)

    def testSetup(self):
        self.assertEqual(len(self.pop.mutations), 2)
        self.assertEqual(len(self.pop.haploid_genomes), 1)
        self.assertEqual(len(self.pop.diploids), 2)
        for g in self.pop.haploid_genomes:
            self.assertEqual(len(g.mutations), 1)
            self.assertEqual(len(g.smutations), 1)
        for mc in self.pop.mcounts:
            self.assertEqual(mc, 4)

    def testWithPruneSelected(self):
        fp.update_mutations(self.pop, 1, 2 * len(self.pop.diploids), True)
        self.assertEqual(len(self.pop.fixations), 2)
        self.assertEqual(len(self.pop.fixation_times), 2)
        self.assertEqual(len([i for i in self.pop.mcounts if i == 0]), 2)
        self.assertTrue(sorted(self.pop.fixations, key=lambda x: x.pos))
        for mc in self.pop.mcounts:
            self.assertEqual(mc, 0)

    def testWithoutPruneSelected(self):
        fp.update_mutations(self.pop, 1, 2 * len(self.pop.diploids), False)
        self.assertEqual(len(self.pop.fixations), 2)
        self.assertEqual(len(self.pop.fixation_times), 2)
        self.assertEqual(len([i for i in self.pop.mcounts if i == 0]), 1)
        self.assertTrue(sorted(self.pop.fixations, key=lambda x: x.pos))
        for m, mc in zip(self.pop.mutations, self.pop.mcounts):
            if m.neutral is True:
                self.assertEqual(mc, 0)
            else:
                self.assertEqual(mc, 4)


class testFixationPreservation(unittest.TestCase):
    @classmethod
    def setUp(self):
        import numpy as np

        N = 1000
        demography = np.array([N] * 10 * N, dtype=np.uint32)
        rho = 1.0
        r = rho / (4 * N)

        a = fwdpy11.Multiplicative(2.0)
        self.p = {
            "nregions": [],
            "sregions": [fwdpy11.ExpS(0, 1, 1, 0.01)],
            "recregions": [fwdpy11.Region(0, 1, 1)],
            "rates": (0.0, 0.00005, r),
            "gvalue": a,
            "demography": demography,
        }
        self.pop = fwdpy11.DiploidPopulation(N)
        self.rng = fwdpy11.GSLrng(101 * 45 * 110 * 210)

    def testPopGenSimWithoutPruning(self):
        import fwdpy11
        import numpy as np

        self.p["prune_selected"] = False
        params = fwdpy11.ModelParams(**self.p)
        fwdpy11.evolve_genomes(self.rng, self.pop, params)
        assert len(self.pop.fixations) > 0, "Test is meaningless without fixations"
        mc = np.array(self.pop.mcounts)
        self.assertEqual(
            len(np.where(mc == 2 * self.pop.N)[0]), len(self.pop.fixations)
        )

    def testPopGenSimWithPruning(self):
        import fwdpy11
        import numpy as np

        self.p["prune_selected"] = True
        params = fwdpy11.ModelParams(**self.p)
        fwdpy11.evolve_genomes(self.rng, self.pop, params)
        assert len(self.pop.fixations) > 0, "Test is meaningless without fixations"
        mc = np.array(self.pop.mcounts)
        self.assertEqual(len(np.where(mc == 2 * self.pop.N)[0]), 0)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import fwdpy11


class testAdditive(unittest.TestCase):
    @classmethod
    def setUp(self):
        GN = fwdpy11.GaussianNoise
        self.w = fwdpy11.Additive(2.0)
        self.t = fwdpy11.Additive(2.0, fwdpy11.GSS(0.0, 1.0))
        self.tn = fwdpy11.Additive(1.0, fwdpy11.GSS(0.0, 1.0), GN(mean=0.1, sd=2.0))

    def testScaling(self):
        self.assertEqual(self.w.scaling, 2.0)
        self.assertEqual(self.t.scaling, 2.0)
        self.assertEqual(self.tn.scaling, 1.0)

    def testFitnessOrTraitBaseClassProperty(self):
        self.assertEqual(self.w.maps_to_fitness, True)
        self.assertEqual(self.w.maps_to_trait_value, False)
        self.assertEqual(self.t.maps_to_fitness, False)
        self.assertEqual(self.t.maps_to_trait_value, True)
        self.assertEqual(self.tn.maps_to_fitness, False)
        self.assertEqual(self.tn.maps_to_trait_value, True)

    def testFitnessOrTrait(self):
        self.assertEqual(self.w.is_fitness, True)
        self.assertEqual(self.t.is_fitness, False)
        self.assertEqual(self.tn.is_fitness, False)

    def testPickleFitness(self):
        import pickle

        p = pickle.dumps(self.w)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.w.scaling)
        self.assertTrue(up.is_fitness)
        self.assertEqual(type(up.noise), type(self.w.noise))
        self.assertEqual(type(up.gvalue_to_fitness), type(self.w.gvalue_to_fitness))

    def testPickleTraitNoNoise(self):
        import pickle

        p = pickle.dumps(self.t)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.t.scaling)
        self.assertTrue(up.is_fitness is False)
        self.assertEqual(type(up.noise), type(self.t.noise))
        self.assertEqual(type(up.gvalue_to_fitness), type(self.t.gvalue_to_fitness))

    def testPickleTraitWithNoise(self):
        import pickle

        p = pickle.dumps(self.tn)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.tn.scaling)
        self.assertTrue(up.is_fitness is False)
        self.assertEqual(type(up.noise), type(self.tn.noise))
        self.assertEqual(type(up.gvalue_to_fitness), type(self.tn.gvalue_to_fitness))

    def testPickleTraitWithNoiseToFile(self):
        import pickle

        with open("ptest.pickle", "wb") as f:
            pickle.dump(self.tn, f)

        with open("ptest.pickle", "rb") as f:
            up = pickle.load(f)
        self.assertEqual(up.scaling, self.tn.scaling)
        self.assertTrue(up.is_fitness is False)
        self.assertEqual(type(up.noise), type(self.tn.noise))
        self.assertEqual(type(up.gvalue_to_fitness), type(self.tn.gvalue_to_fitness))


class testMultiplicative(unittest.TestCase):
    @classmethod
    def setUp(self):
        GN = fwdpy11.GaussianNoise
        self.w = fwdpy11.Multiplicative(2.0)
        self.t = fwdpy11.Multiplicative(2.0, fwdpy11.GSS(0.0, 1.0))
        self.tn = fwdpy11.Multiplicative(
            1.0, fwdpy11.GSS(0.0, 1.0), GN(mean=0.1, sd=2.0)
        )

    def testScaling(self):
        self.assertEqual(self.w.scaling, 2.0)
        self.assertEqual(self.t.scaling, 2.0)
        self.assertEqual(self.tn.scaling, 1.0)

    def testFitnessOrTrait(self):
        self.assertEqual(self.w.is_fitness, True)
        self.assertEqual(self.t.is_fitness, False)
        self.assertEqual(self.tn.is_fitness, False)

    def testFitnessOrTraitBaseClassProperty(self):
        self.assertEqual(self.w.maps_to_fitness, True)
        self.assertEqual(self.w.maps_to_trait_value, False)
        self.assertEqual(self.t.maps_to_fitness, False)
        self.assertEqual(self.t.maps_to_trait_value, True)
        self.assertEqual(self.tn.maps_to_fitness, False)
        self.assertEqual(self.tn.maps_to_trait_value, True)

    def testPickleFitness(self):
        import pickle

        p = pickle.dumps(self.w)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.w.scaling)
        self.assertTrue(up.is_fitness)
        self.assertEqual(type(up.noise), type(self.w.noise))
        self.assertEqual(type(up.gvalue_to_fitness), type(self.w.gvalue_to_fitness))

    def testPickleTraitNoNoise(self):
        import pickle

        p = pickle.dumps(self.t)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.t.scaling)
        self.assertTrue(up.is_fitness is False)
        self.assertEqual(type(up.noise), type(self.t.noise))
        self.assertEqual(type(up.gvalue_to_fitness), type(self.t.gvalue_to_fitness))

    def testPickleTraitWithNoise(self):
        import pickle

        p = pickle.dumps(self.tn)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.tn.scaling)
        self.assertTrue(up.is_fitness is False)
        self.assertEqual(type(up.noise), type(self.tn.noise))
        self.assertEqual(type(up.gvalue_to_fitness), type(self.tn.gvalue_to_fitness))


class testGBR(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.gss = fwdpy11.GSS(0.0, 1.0)
        self.gnoise = fwdpy11.GaussianNoise(mean=0.0, sd=1.0)
        self.nonoise = fwdpy11.NoNoise()

    def testPicklingGSS(self):
        import pickle

        gbr = fwdpy11.GBR(self.gss)
        self.assertFalse(gbr.maps_to_fitness)
        self.assertFalse(gbr.maps_to_fitness)
        p = pickle.dumps(gbr, -1)
        up = pickle.loads(p)
        self.assertEqual(type(self.nonoise), type(up.noise))
        self.assertEqual(type(self.gss), type(up.gvalue_to_fitness))

    def testPicklingGSSGaussianNoise(self):
        import pickle

        gbr = fwdpy11.GBR(self.gss, self.gnoise)
        self.assertFalse(gbr.maps_to_fitness)
        self.assertFalse(gbr.maps_to_fitness)
        p = pickle.dumps(gbr, -1)
        up = pickle.loads(p)
        self.assertEqual(type(self.gnoise), type(up.noise))
        self.assertEqual(type(self.gss), type(up.gvalue_to_fitness))


class testGSS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.x = fwdpy11.GSS(0.0, 1.0)

    def testProperties(self):
        self.assertEqual(self.x.opt, 0.0)
        self.assertEqual(self.x.VS, 1.0)


class testGSSandGSSmoConsistency(unittest.TestCase):
    """
    This tests that GSS and GSSmo have opt and VS
    in the same order.  If that were not true,
    fitness calculations would come out differently
    and the test would fail.
    """

    @classmethod
    def setUp(self):
        self.a = fwdpy11.Additive(2.0, fwdpy11.GSS(0.0, 1.0))
        self.b = fwdpy11.Additive(2.0, fwdpy11.GSSmo([(0, 0.0, 1.0)]))
        self.pop = fwdpy11.DiploidPopulation(1000)

    def testFitnesses(self):
        wa = [self.a.fitness(i, self.pop) for i in range(self.pop.N)]
        wb = [self.b.fitness(i, self.pop) for i in range(self.pop.N)]
        for i, j in zip(wa, wb):
            self.assertEqual(i, j)


if __name__ == "__main__":
    unittest.main()
